"""
This package contains command-line tools that will be installed as `auromat-name`
for a module name `auromat.cli.name`. It is not intended to be used from within
Python code. The existing command-line tool modules may be used as references
to develop new special-purpose command-line tools.
"""